@@ -47,12 +47,18 @@
 (cl-defstruct llm-ollama
   "A structure for holding information needed by Ollama's API.
 
+SCHEME is the http scheme to use, a string. It is optional and
+default to `http'.
+
+HOST is the host that Ollama is running on. It is optional and
+default to localhost.
+
 PORT is the localhost port that Ollama is running on.  It is optional.
 
 CHAT-MODEL is the model to use for chat queries. It is required.
 
 EMBEDDING-MODEL is the model to use for embeddings.  It is required."
-  port chat-model embedding-model)
+  (scheme "http") (host "localhost") (port 11434) chat-model embedding-model)
 
 ;; Ollama's models may or may not be free, we have no way of knowing. There's no
 ;; way to tell, and no ToS to point out here.
@@ -62,7 +68,8 @@ EMBEDDING-MODEL is the model to use for embeddings.  It is required."
 
 (defun llm-ollama--url (provider method)
   "With ollama PROVIDER, return url for METHOD."
-  (format "http://localhost:%d/api/%s" (or (llm-ollama-port provider) 11434) method))
+  (format "%s://%s:%d/api/%s" (llm-ollama-scheme provider )(llm-ollama-host provider)
+          (llm-ollama-port provider) method))
 
 (defun llm-ollama--embedding-request (provider string)
   "Return the request to the server for the embedding of STRING.
@@ -75,14 +82,17 @@ PROVIDER is the llm-ollama provider."
   (assoc-default 'embedding response))
 
 (cl-defmethod llm-embedding-async ((provider llm-ollama) string vector-callback error-callback)
-  (llm-request-async (llm-ollama--url provider "embeddings")
+  (let ((buf (current-buffer)))
+    (llm-request-async (llm-ollama--url provider "embeddings")
                      :data (llm-ollama--embedding-request provider string)
                      :on-success (lambda (data)
-                                   (funcall vector-callback (llm-ollama--embedding-extract-response data)))
+                                   (llm-request-callback-in-buffer
+                                    buf vector-callback (llm-ollama--embedding-extract-response data)))
                      :on-error (lambda (_ _)
                                  ;; The problem with ollama is that it doesn't
                                  ;; seem to have an error response.
-                                 (funcall error-callback 'error "Unknown error calling ollama"))))
+                                 (llm-request-callback-in-buffer
+                                  buf error-callback 'error "Unknown error calling ollama")))))
 
 (cl-defmethod llm-embedding ((provider llm-ollama) string)
   (llm-ollama--embedding-extract-response
@@ -106,20 +116,16 @@ STREAMING if non-nil, turn on response streaming."
                                          (car example)
                                          (cdr example)))
                                (llm-chat-prompt-examples prompt) "\n"))))
-    (setq text-prompt (concat text-prompt "\n"
-                              (let ((conversationp (> (length (llm-chat-prompt-interactions prompt)) 1)))
-                                (if conversationp
-                                    (concat
-                                     "The following interactions have already happened: "
-                                     (mapcar (lambda (p)
-                                               (format "%s: %s\n"
-                                                       (pcase (llm-chat-prompt-interaction-role p)
-                                                         ('user "User")
-                                                         ('assistant "Assistant"))
-                                                       (string-trim (llm-chat-prompt-interaction-content p))))
-                                             (llm-chat-prompt-interactions prompt)))
-                                  (string-trim
-                                   (llm-chat-prompt-interaction-content (car (llm-chat-prompt-interactions prompt))))))))
+    ;; The last item always should be the latest interaction, which is the prompt.
+    (setq text-prompt (concat text-prompt
+                              "\n"
+                              (string-trim (llm-chat-prompt-interaction-content
+                                            (car (last (llm-chat-prompt-interactions prompt)))))))
+    ;; If the first item isn't an interaction, then it's a conversation which
+    ;; we'll set as the chat context.
+    (when (not (eq (type-of (car (llm-chat-prompt-interactions prompt)))
+                   'llm-chat-prompt-interaction))
+      (push `("context" . ,(car (llm-chat-prompt-interactions prompt))) request-alist))
     (push `("prompt" . ,(string-trim text-prompt)) request-alist)
     (push `("model" . ,(llm-ollama-chat-model provider)) request-alist)
     (when (llm-chat-prompt-temperature prompt)
@@ -132,55 +138,85 @@ STREAMING if non-nil, turn on response streaming."
 (defvar-local llm-ollama-current-response ""
   "The response so far from the server.")
 
-(defvar-local llm-ollama-last-position 1
-  "The last position in the streamed response we read until.")
+(defvar-local llm-ollama-last-response 0
+  "The last response number we've read.")
 
 (defun llm-ollama--get-partial-chat-response (response)
   "Return the text in the partial chat response from RESPONSE."
   ;; To begin with, we should still be in the buffer with the actual response.
   (let ((current-response llm-ollama-current-response)
-        (last-position llm-ollama-last-position))
+        (last-response llm-ollama-last-response))
     (with-temp-buffer
       (insert response)
-      (goto-char last-position)
-      (while (search-forward "{" nil t)
-        (backward-char 1)
-        (ignore-errors
-          (let ((obj (json-read)))
-            (unless (eq (assoc-default 'done obj) :json-true)
-              (setq current-response
-                    (concat current-response (assoc-default 'response obj))))))
-        (setq last-position (point))))
-    (setq-local llm-ollama-current-response current-response)
-    (setq-local llm-ollama-last-position last-position)
+      ;; Responses in ollama are always one per line.
+      (let* ((end-pos (save-excursion (goto-char (point-max))
+                                      (when (search-backward-regexp
+                                             (rx (seq "done\":false}" line-end))
+                                             nil t)
+                                        (pos-eol)))))
+        (when end-pos
+          (let ((all-lines (seq-filter
+                            (lambda (line) (string-match-p (rx (seq string-start ?{)) line))
+                            (split-string (buffer-substring-no-properties 1 end-pos) "\n" t))))
+            (setq
+             current-response
+             (concat current-response
+                     (mapconcat
+                      (lambda (line) (assoc-default 'response (json-read-from-string line)))
+                      ;; Take from response output last-response to the end. This
+                      ;; counts only valid responses, so we need to throw out all
+                      ;; other lines that aren't valid JSON.
+                      (seq-subseq all-lines last-response) "")))
+            (setq last-response (length all-lines))))))
+    ;; If there is no new content, don't manipulate anything.
+    (when (> (length current-response) (length llm-ollama-current-response))
+      (setq llm-ollama-last-response last-response)
+      (setq llm-ollama-current-response current-response))
     current-response))
 
+(defun llm-ollama--get-final-response (response)
+  "Return the final post-streaming json output from RESPONSE."
+  (with-temp-buffer
+    (insert response)
+    ;; Find the last json object in the buffer.
+    (goto-char (point-max))
+    (search-backward "{" nil t)
+    (json-read)))
+
 (cl-defmethod llm-chat ((provider llm-ollama) prompt)
   ;; We expect to be in a new buffer with the response, which we use to store
   ;; local variables. The temp buffer won't have the response, but that's fine,
   ;; we really just need it for the local variables.
   (with-temp-buffer
-    (llm-ollama--get-partial-chat-response
-     (llm-request-sync-raw-output (llm-ollama--url provider "generate")
-                                  :data (llm-ollama--chat-request provider prompt)
-                                  ;; ollama is run on a user's machine, and it can take a while.
-                                  :timeout llm-ollama-chat-timeout))))
+    (let ((output (llm-request-sync-raw-output 
+                   (llm-ollama--url provider "generate")
+                   :data (llm-ollama--chat-request provider prompt)
+                   ;; ollama is run on a user's machine, and it can take a while.
+                   :timeout llm-ollama-chat-timeout)))
+      (setf (llm-chat-prompt-interactions prompt)
+	        (list (assoc-default 'context (llm-ollama--get-final-response output))))
+      (llm-ollama--get-partial-chat-response output))))
 
 (cl-defmethod llm-chat-async ((provider llm-ollama) prompt response-callback error-callback)
   (llm-chat-streaming provider prompt (lambda (_)) response-callback error-callback))
 
 (cl-defmethod llm-chat-streaming ((provider llm-ollama) prompt partial-callback response-callback error-callback)
-  (llm-request-async (llm-ollama--url provider "generate")
+  (let ((buf (current-buffer)))
+    (llm-request-async (llm-ollama--url provider "generate")
       :data (llm-ollama--chat-request provider prompt)
-      :on-success-raw (lambda (data)
-                        (funcall response-callback (llm-ollama--get-partial-chat-response data)))
+      :on-success-raw (lambda (response)
+                        (setf (llm-chat-prompt-interactions prompt)
+                              (list (assoc-default 'context (llm-ollama--get-final-response response))))
+                        (llm-request-callback-in-buffer
+                         buf response-callback
+                         (llm-ollama--get-partial-chat-response response)))
       :on-partial (lambda (data)
                     (when-let ((response (llm-ollama--get-partial-chat-response data)))
-                      (funcall partial-callback response)))
+                      (llm-request-callback-in-buffer buf partial-callback response)))
       :on-error (lambda (_ _)
                   ;; The problem with ollama is that it doesn't
                   ;; seem to have an error response.
-                  (funcall error-callback 'error "Unknown error calling ollama"))))
+                  (llm-request-callback-in-buffer buf error-callback "Unknown error calling ollama")))))
 
 (provide 'llm-ollama)
 
